@@ -28,6 +28,12 @@ except ImportError:
     from pymatgen.core.lattice import Lattice   
 from pymatgen.analysis.structure_matcher import StructureMatcher
 from scipy import spatial
+from scipy.sparse import coo_matrix
+
+import warnings
+
+# Used for faster version of EMD
+import numba as nb
 
 # the wasserstein distance in scipy treats frequencies of each bin as a value, 
 # and then builds the distributions from those values and computes the distance. 
@@ -47,6 +53,7 @@ from pyemd import emd, emd_with_flow
 
 from .data_io import rdf_read, rdf_read_parallel
 from .data_explore import rdf_trim, rdf_flatten
+import gridrdf.composition
 from .composition import composition_one_hot, element_indice
 from .misc import int_or_str
 
@@ -68,15 +75,15 @@ def emd_formula_example():
     comp2 = elem_emd._gen_vector('La0.57Li0.29TiO3')
     pettifor_emd = elem_emd._EMD(comp1, comp2)
 
-    comp1_reindex = pd.DataFrame(comp1, index=modified_pettifor).reindex(index=pettifor)
-    comp2_reindex = pd.DataFrame(comp2, index=modified_pettifor).reindex(index=pettifor)
+    comp1_reindex = pd.DataFrame(comp1, index=gridrdf.composition.modified_pettifor).reindex(index=gridrdf.composition.pettifor)
+    comp2_reindex = pd.DataFrame(comp2, index=gridrdf.composition.modified_pettifor).reindex(index=gridrdf.composition.pettifor)
 
     dist_matrix = pd.read_csv('similarity_matrix.csv', index_col='ionA').values
     dist_matrix = dist_matrix.copy(order='C')
 
     em = emd_with_flow(comp1_reindex.values[:,0], comp2_reindex.values[:,0], dist_matrix)
     simi_matrix_emd = em[0]
-    emd_flow = pd.DataFrame(em[1], columns=pettifor, index=pettifor)
+    emd_flow = pd.DataFrame(em[1], columns=gridrdf.composition.pettifor, index=gridrdf.composition.pettifor)
     emd_flow.replace(0, np.nan).to_csv('emd_flow.csv')
 
 
@@ -188,6 +195,8 @@ def dist_matrix_1d(nbin=100):
     '''
     Generate distance matrix for earth mover's distance use
     '''
+
+    #dist_matrix = np.abs(np.subtract.outer(range(nbin), range(nbin)))
     dist_matrix = pd.DataFrame()
     for x in range(nbin):
         for y in range(nbin):
@@ -275,6 +284,9 @@ def nn_bulk_modulus_matrix_add(data, nn=1, simi_dir = '.', simi_matrix=['extende
         EMD distance between the nearest neighbor and baseline
         mp-id of the nearest neighbor
     '''
+
+    warnings.warn('`nn_bulk_modulus_matrix_add` is deprecated and will be removed', DeprecationWarning)
+
     # read the modulus data and convert to pandas dataframe
     # for the usage of later ground truth and mp ids
     df_mp = pd.DataFrame.from_dict(data)
@@ -416,7 +428,7 @@ def composition_similarity(baseline_id, data, index='z_number_78'):
     dist_matrix = 1 / (np.log10(1 / dist_matrix + 1))
 
     if index == 'pettifor':
-        dist_matrix = dist_matrix.reindex(columns=pettifor, index=pettifor) 
+        dist_matrix = dist_matrix.reindex(columns=gridrdf.composition.pettifor, index=gridrdf.composition.pettifor) 
     # the earth mover's distance package need order C and float64 astype('float64')
     dist_matrix = dist_matrix.values.copy(order='C')
 
@@ -429,7 +441,7 @@ def composition_similarity(baseline_id, data, index='z_number_78'):
     return compo_emd
 
 
-def composition_similarity_matrix(data, indice=None, index='z_number_78'):
+def composition_similarity_matrix(data, indice=None, index='z_number_78', elem_similarity_file='similarity_matrix.csv'):
     '''
     Calcalate pairwise earth mover's distance of all compositions, the composition should be 
     a 78-element vector, as the elemental similarity_matrix is 78x78 matrix in the order of  
@@ -455,16 +467,16 @@ def composition_similarity_matrix(data, indice=None, index='z_number_78'):
         indice = [0, len(data)]
     
     dist = []
-    elem_similarity_file = os.path.join(sys.path[0], 'similarity_matrix.csv')
     dist_matrix = pd.read_csv(elem_similarity_file, index_col='ionA')
     dist_matrix = 1 / (np.log10(1 / dist_matrix + 1))
 
     if index == 'pettifor':
-        dist_matrix = dist_matrix.reindex(columns=pettifor, index=pettifor) 
+        dist_matrix = dist_matrix.reindex(columns=gridrdf.composition.pettifor, index=gridrdf.composition.pettifor) 
     # the earth mover's distance package need order C and float64 astype('float64')
     dist_matrix = dist_matrix.values.copy(order='C')
-
-    compo_emd = pd.DataFrame([])
+    
+    ids = data.iloc[indice[0]:indice[1]].index
+    compo_emd = pd.DataFrame(np.zeros((indice[-1], indice[-1])), index=ids, columns=ids )
     for i1 in range(indice[0], indice[1]):
         mp_id_1 = data.index[i1]
         for i2, mp_id_2 in enumerate(data.index):
@@ -477,8 +489,245 @@ def composition_similarity_matrix(data, indice=None, index='z_number_78'):
                 compo_emd.loc[mp_id_1, mp_id_2] = np.nan
     return compo_emd
 
+@nb.njit(parallel=True)
+def _emd_cumsum_row(cumsum_grid_array, index, bin_width, weights = None):
+    """Super-fast implementation of EMD with Numba support for an array of structures.
+    
+    WARNING: This algorithm does not do any error checking! It assumes that the input is 
+             a NumPy array of cumulative, normalised grid distributions.
 
-def rdf_similarity(baseline_rdf, all_rdf):
+    This is optimised to calculate EMD between one GRID and many others in parallel, with
+    the aim to create a 2D EMD distance matrix. As such, it only computes EMD to structures
+    with index greater or equal to `index`.
+
+
+    Parameters
+    ----------
+    
+    cumsum_grid_array: np.ndarray, shape (nstructures, ngrid_shells, nbins)
+        Cumulative distributions of normalised GRIDs, stacked into a single 3D tensor. 
+        IMPORTANT: cumulative distributions must all sum to the same value!
+
+    index: int
+        Index of structure to calculate pairwise EMDs for
+    bin_width : float
+        Width of each histogram bin, used to normalise the resulting EMD correctly
+    weights: array-like, default None
+        Weighting to apply to each grid shell before computing the mean. If None, equal weighting (1/n_shells) 
+        will be applied.
+        The weights should be the same length as the number of GRID shells, and should sum to the number of shells (not 1).
+
+
+    Notes
+    -----
+
+    As detailed [scipy.stats.wasserstein_distance](https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.wasserstein_distance.html), 
+    an equivalent definition of EMD based on cumulative distribution functions is:
+
+    $$
+    EMD(u,v) = \int_{-\infty}{+\infty}|U - V|
+    $$
+    where $u,v$ are probability distributions and $U, V$ are their respectives cumulative distributions. As such, the EMD between 
+    two GRID representations $(G,H)$ will be
+    $$
+    EMD(G, H) = \sum_{j=1}^n \left( \frac{\sum_i | G_{j,i} - H_{j,i} |}{max(U)} \times w \right) / n
+    $$
+    where the sum over $i$ is for all bins in a given GRID shell (to the same cutoff) and $w$ is the width of each bin. The 
+    division by $max(U)$ is needed in cases where $u$ and $v$ are not normalised to sum to 1, for instance when $w < 1$.
+    $j$ is the sum across all GRID shells $n$, meaning that the overall EMD is the mean of each shell EMD (currently, this
+    may change in future).
+
+    """
+
+    if weights is None:
+        shell_weights = np.ones(cumsum_grid_array.shape[1])
+    else:
+        # Avoid numba type errors by simply multiplying ones by weights (also checks array sizes)
+        shell_weights = np.ones(cumsum_grid_array.shape[1]) * weights
+
+    output = np.zeros(cumsum_grid_array.shape[0])
+    for j in nb.prange(index,cumsum_grid_array.shape[0]):
+        output[j] = np.mean( shell_weights * np.sum(np.abs(cumsum_grid_array[index] - cumsum_grid_array[j]), axis=-1) / np.max(cumsum_grid_array[index]) * bin_width)
+        #output[j] = np.sum(shell_weights)
+    return output
+
+@nb.njit()
+def _EMD_cumsum(cumulative_grid1, cumulative_grid2, bin_width, weights=None):
+    """
+    Return EMD between two cumulative GRID (2D) representations.
+
+    Parameters
+    ----------
+
+    cumulative_grid1, cumulative_grid2: array-like
+        2D arrays containing cumulative distribution values with which to calculate Earth Mover's Distance
+    bin_width: float
+        Width of each bin in order to normalise the EMD correctly
+    weights: array-like, default None
+        Weighting to apply to each grid shell before computing the mean. If None, equal weighting (1/n_shells) 
+        will be applied.
+        The weights should be the same length as the number of GRID shells, and should sum to 1.
+
+    Returns
+    -------
+
+    EMD : float
+        Earth mover distance between two GRID distributions
+
+    Notes
+    -----
+
+    For mathematical explanation see `_EMD_cumsum_row`
+
+    """
+    
+    if weights is None:
+        weights = np.ones(cumulative_grid1.shape[0])
+    return np.mean( weights * np.sum(np.abs(cumulative_grid1 - cumulative_grid2), axis=-1) / np.max(cumulative_grid1) * bin_width)
+
+@nb.njit()
+def _flattened_EMD_cumsum(cumulative_grid_flat1, cumulative_grid_flat2, shape, bin_width, weights=None):
+    """
+    Compute EMD between two flattened GRID distributions.
+    """
+
+    grid1 = cumulative_grid_flat1.reshape(shape)
+    grid2 = cumulative_grid_flat2.reshape(shape)
+
+    return _EMD_cumsum(grid1, grid2, bin_width, weights)
+
+def super_fast_EMD_matrix(grids,
+                          bin_width,
+                          weighting='constant',
+                          weighting_kwargs={'n':0.0},
+                          results_array = None,):
+    """
+    Efficient parallelised method to compute the matrix of EMD for a collection of GRID representations.
+
+    Warning - for large lists of GRIDS, this can be very memory (and compute) intensive!
+
+    Parameters
+    ----------
+    grids : array-like
+        Iterable of GRID representations
+    bin_width : float
+        Width of each GRID bin for correctly normalising EMD
+    weighting : str or function
+        Weighting to apply to each grid shell when computing the overall EMD, allowing
+        biasing towards short- or long-range similarity. A string argument will use 
+        a pre-defined weighting function, or alternatively a callable can be provided in
+        order to compute weightings. If callable, the function should take the number of 
+        shells as input and return an array of values.
+        Note that weights should be normalised to sum to the number
+        of GRID shells (not 1).
+        In-built options:
+            'constant': equal weighting to all GRID shells
+            'power' : weighting of shell x is 1/x**n (normalised over all shells)
+        Default : 'constant' (equal weighting to bins)
+    weighting_kwargs : dict
+        Kwargs to use for computing weighting, either as parameters to in-built 
+        functions or to be passed direct to a weighting callable.
+    results_array : array-like, default None
+        If not None, EMD values will be stored to this array. It should
+        be square with shape (len(grids), len(grids)).
+        Can be substituted with an HDF5 file object to minimise memory consumption.
+
+
+    Notes
+    -----
+
+    - This method requires all GRIDs to be the same shape so that they can be converted into a 3-dimensional NumPy array.
+    - Checks are made for normalisation
+    """
+
+    shapes = np.array([i.shape for i in grids])
+    assert np.all(np.isclose(shapes, shapes[0])), "All GRIDs must have the same dimensions"
+
+    if weighting == 'constant':
+        # Use a constant weighting
+        weights = np.ones(grids[0].shape[0])
+    elif weighting == 'power':
+        # Use x**(-n)
+        assert 'n' in weighting_kwargs, "weighting_kwargs must contain `n` when weighting==power"
+        power_series = np.power(np.arange(1, grids[0].shape[0]+1), -weighting_kwargs['n'])
+        # Normalise the weights to n_grid_shells
+        weights = power_series * grids[0].shape[0] / power_series.sum()
+    elif callable(weighting):
+        weights = weighting(grids[0].shape[0], **weighting_kwargs)
+
+
+    if not isinstance(grids, np.ndarray):
+        grid_input = np.array(grids)
+    else:
+        grid_input = grids
+
+    # Perform cumulative summation for fast EMD calculation
+    grid_cumsum = np.cumsum(grid_input, axis=-1)
+
+    assert np.all(np.isclose(grid_cumsum[0,0,-1], grid_cumsum[:,:,-1])), "All GRIDs must be normalised to the same area. "
+
+    if results_array:
+        assert results_array.shape == (grid_input.shape[0], grid_input.shape[0])
+        output = results_array
+    else:
+        output = np.zeros((grid_cumsum.shape[0], grid_cumsum.shape[0]))
+
+    # Do the time-consuming calculation in parallel (with progress bar)
+    for i in tqdm(range(grid_cumsum.shape[0]), mininterval=5):
+        output[i] = _emd_cumsum_row(grid_cumsum, i, bin_width, weights=weights)
+        # Fill in the lower diagonal (assumes we are iterating in order)
+        output[i, :i] = output[:i, i]
+    return output
+
+def rdf_emd_similarity(rdf_a, rdf_b, max_distance=10, method='fast'):
+    """
+    Compute EMD similarity between two RDF distributions (1D or 2D)
+    
+    Notes
+    -----
+    
+    This is the most time-consuming step in computing EMD similarity between distributions (particularly GRIDs).
+    TODO: optimise this method further
+    
+    The "fast" method joins all GRID shells into a single 1D vector, and uses a modified distance metric so that
+    comparisons between equivalent shells will give the same EMD, but distributions between shells have a high EMD contribution
+    (**unless the distributions are not normalised**). This requires only one call to `wasserstein_distance` (with n_shell*max_dist bins)
+    rather than n_shell calls to `wasserstein_distance` (each with max_dist bins).
+    
+    """
+    
+    assert rdf_a.shape == rdf_b.shape
+    emd_bins = np.linspace(0, max_distance, rdf_a.shape[-1])
+    
+    if len(rdf_a.shape) == 2:
+        if method=='orig':
+            shell_distances = []
+            for j in range(rdf_a.shape[0]):
+                shell_distances.append(wasserstein_distance(emd_bins, emd_bins, 
+                                                            rdf_a[j], rdf_b[j]))
+                                                            
+
+            return np.array(shell_distances).mean()
+        elif method=='fast':
+            assert np.isclose(rdf_a.sum(axis=1), 1.0).all(), "All GRID shells must be normalized to 1.0 in order for the fast method to work correctly."
+            assert np.isclose(rdf_b.sum(axis=1), 1.0).all(), "All GRID shells must be normalized to 1.0 in order for the fast method to work correctly."
+            shell_offsets = np.arange(0, rdf_a.shape[0]) * emd_bins[-1]
+            emd_bins_1D = (np.vstack([emd_bins]*rdf_a.shape[0]) + np.column_stack([shell_offsets] * rdf_a.shape[-1]))
+            if isinstance(rdf_a, np.ndarray) and isinstance(rdf_b, np.ndarray):
+                # Using standard NumPy array processing
+                emd_bins_1D = emd_bins_1D.ravel()
+                return wasserstein_distance(emd_bins_1D, emd_bins_1D, rdf_a.ravel(), rdf_b.ravel())
+            elif isinstance(rdf_a, coo_matrix) and isinstance(rdf_b, coo_matrix):
+                # We are using sparse array, so can gain a significant speedup (~2-5-fold)
+                emd_bins_a = emd_bins_1D[rdf_a.row, rdf_a.col]
+                emd_bins_b = emd_bins_1D[rdf_b.row, rdf_b.col]
+                return wasserstein_distance(emd_bins_a, emd_bins_b, rdf_a.data, rdf_b.data)
+    elif len(rdf_a.shape) == 1:
+        return wasserstein_distance(emd_bins, emd_bins, rdf_a, rdf_b)
+                                                            
+    
+    
+def rdf_row_similarity(baseline_rdf, all_rdf, max_distance = 10):
     '''
     Calculate the earth mover's distance between two RDFs
     Current support vanilla rdf and extend rdf
@@ -500,12 +749,13 @@ def rdf_similarity(baseline_rdf, all_rdf):
     if len(rdf_1.shape) == 2:
         dist_matrix = dist_matrix_1d(len(rdf_1[0]))
         for d, rdf_2 in enumerate(all_rdf):
-            rdf_len = np.array([len(rdf_1), len(rdf_2)]).min()
-            shell_distances = []
-            for j in range(rdf_len):
-                shell_distances.append(wasserstein_distance(emd_bins, emd_bins, rdf_1[j], rdf_2[j]))
-                #shell_distances.append(emd(rdf_1[j], rdf_2[j], dist_matrix))
-            rdf_emd.loc[d] = np.array(shell_distances).mean()
+            #rdf_len = np.array([len(rdf_1), len(rdf_2)]).min()
+            #shell_distances = []
+            #for j in range(rdf_len):
+            #    shell_distances.append(wasserstein_distance(emd_bins, emd_bins, rdf_1[j], rdf_2[j]))
+            #    #shell_distances.append(emd(rdf_1[j], rdf_2[j], dist_matrix))
+            #rdf_emd.loc[d] = np.array(shell_distances).mean()
+            rdf_emd.loc[d] = rdf_emd_similarity(rdf_1, rdf_2, max_distance=max_distance)
 
     elif len(rdf_1.shape) == 1:
         dist_matrix = dist_matrix_1d(len(rdf_1))
@@ -516,63 +766,70 @@ def rdf_similarity(baseline_rdf, all_rdf):
     return rdf_emd
 
 
-def rdf_similarity_matrix(data,
-                          all_rdf,
+def rdf_similarity_matrix(all_rdf,
+                          data=None,
                           indice=None,
                           method='emd',
                           max_distance = 10,
                           ):
     '''
-    Calculate the earth mover's distance between all RDF pairs in a large dataset
-    Scipy wassertein is used
-    Using Guassian smearing for rdf
+    Calculate the earth mover's distance between all RDF pairs, returning a square matrix.
 
-    Args:
-        data: data from json
-        all_rdf:
-        indice:start and end of the index of the dataset, this parameter is useful when 
-            the dataset is large, and the calculations of the whole matrix must be splitted
-            into different tasks
-    Return:
-        a pandas dataframe with all pairwise distance
-        for multiple shells rdf the distance is the mean value of all shells
+    Parameters
+    ----------
+
+    all_rdf : array-like
+        Iterable of RDF representations to use for RDF calculation
+    data: iterable of dicts, optional
+        Iterable the same length as all_rdf, containing the key `task_id` to be used to
+        label the resulting DataFrame.
+
+    Returns
+    -------
+
+    dissimilarity : DataFrame
+        DataFrame containing pairwise dissimilarity between RDF representations.
+
+    Notes
+    -----
+
+    `rdf_similarity_matrix` has changed to use `super_fast_EMD_matrix` internally, which is better
+    parallelised. More control over data storage (e.g. saving direct to a file to reduce memory) can
+    be achieved by using `super_fast_EMD_matrix` directly.
     '''
     
+    warnings.warn('`rdf_similarity` is now effectively parallelised using numba, so indice is no longer used.', SyntaxWarning)
+
     rdf_lengths = [i.shape[0] for i in all_rdf]
     assert len(set(rdf_lengths)) == 1, "All RDFs (or GRIDs) must have the same number of bins - use `data_prepare.trim_rdf_bins` first"
     
-    # used for wasserstein distance.
-    emd_bins = np.linspace(0, max_distance, all_rdf[0].shape[-1])
-
-    # if indice is None, then loop over the whole dataset
-    if not indice:
-        indice = [0, len(data)]
+    # Compute bin_width based on max distance
+    bin_width = max_distance / rdf_lengths[0]
 
     if len(all_rdf[0].shape) == 2:
         # typically for extend RDF
         
-        ids = []
-        for i in range(indice[0], indice[1]):
-            ids.append(data[i]['task_id'])
-        df = pd.DataFrame(np.zeros((len(ids), len(ids))), index=ids, columns=ids)
-        for i1 in tqdm(range(indice[0], indice[1]), desc='similarity matrix rows', mininterval=10):
-            d1 = data[i1]
-            for i2, d2 in enumerate(data):
-                if i1 <= i2:
-                    shell_distances = []
-                    for j in range(len(all_rdf[0])):
-                        if method == 'emd':
-                            try:
-                                shell_distances.append(wasserstein_distance(emd_bins, emd_bins, 
-                                                                        all_rdf[i1][j], all_rdf[i2][j]))
-                            except ValueError:
-                                print('Error determining distance for {}; consider removing from analysis'.format(i1))
-                        elif method == 'cosine':
-                            shell_distances.append(spatial.distance.cosine(all_rdf[i1][j], all_rdf[i2][j]))
-                        
-                    df.loc[d1['task_id'], d2['task_id']] = np.array(shell_distances).mean()
-                else:
-                    df.loc[d1['task_id'], d2['task_id']] = np.nan
+        if data:
+            ids = [i['task_id'] for i in data]
+        else:
+            ids = np.arange(len(all_rdf))
+        
+        dissimilarity = super_fast_EMD_matrix(all_rdf,
+                                              bin_width,
+                                              weighting='constant',
+                                              weighting_kwargs={'n':0.0},
+                                              results_array = None,)
+        
+        df = pd.DataFrame(dissimilarity, index=ids, columns=ids)
+
+        # df = pd.DataFrame(np.zeros((len(ids), len(ids))), index=ids, columns=ids)
+        # for i1 in tqdm(range(indice[0], indice[1]), desc='similarity matrix rows', mininterval=10):
+        #     d1 = data[i1]
+        #     for i2, d2 in enumerate(data):
+        #         if i1 <= i2:
+        #             df.loc[d1['task_id'], d2['task_id']] = rdf_emd_similarity(all_rdf[i1], all_rdf[i2], max_distance=max_distance)
+        #         else:
+        #             df.loc[d1['task_id'], d2['task_id']] = np.nan
     elif len(all_rdf[0].shape) == 1:
         # typically for vanilla RDF and other 1D input
         df = pd.DataFrame([])
@@ -581,8 +838,8 @@ def rdf_similarity_matrix(data,
             for i2, d2 in enumerate(data):
                 if i1 <= i2:
                     if method == 'emd':
-                        shell_distance = wasserstein_distance(emd_bins, emd_bins, 
-                                                            all_rdf[i1], all_rdf[i2])
+                        shell_distance = rdf_emd_similarity(all_rdf[i1], all_rdf[i2], max_distance=max_distance)
+                        
                     elif method == 'cosine':
                         shell_distance = spatial.distance.cosine(all_rdf[i1], all_rdf[i2])
 
@@ -622,6 +879,9 @@ def rdf_similarity_matrix_old(data, all_rdf, order=None, method='emd'):
         a pandas dataframe with all pairwise distance
         for multiple shells rdf the distance is the mean value of all shells
     '''
+
+    warnings.warn("`rdf_similarity_matrix_old` is deprecated, and will be removed in a future release.", DeprecationWarning)
+
     # used for wasserstein distance
     emd_bins = np.linspace(0, 10, 101)
 
@@ -818,7 +1078,7 @@ if __name__ == '__main__':
         else:
             all_rdf = rdf_read(data, rdf_dir)
         baseline_rdf = np.loadtxt(rdf_dir + '/' + baseline_id, delimiter=' ')
-        rdf_emd = rdf_similarity(baseline_rdf, all_rdf)
+        rdf_emd = rdf_row_similarity(baseline_rdf, all_rdf)
         rdf_emd.to_csv(output_file + baseline_id + '_rdf_emd.csv')
 
     elif task == 'rdf_similarity_matrix':
@@ -826,14 +1086,9 @@ if __name__ == '__main__':
             all_rdf = rdf_read_parallel(data, rdf_dir, procs=procs)
         else:
             all_rdf = rdf_read(data, rdf_dir)
-        # trim all the rdf to same length to save time
-        rdf_len = 100
-        all_rdf = rdf_trim(all_rdf, trim=rdf_len)
-        df = rdf_similarity_matrix(data, all_rdf, indice=indice, method='emd')
-        if indice:
-            df.to_csv(output_file + str(indice[0]) + '_' + str(indice[1]) + '.csv')
-        else:
-            df.to_csv(output_file + '_whole_matrix.csv')
+
+        df = rdf_similarity_matrix(all_rdf, data, indice=indice, method='emd')
+        df.to_csv(output_file + '_whole_matrix.csv')
 
     elif task == 'composition_similarity':
         elem_vectors, elem_symbols = composition_one_hot(data)
